use std::ops::Add;

use git2::{ObjectType, Oid, Repository, BranchType};

use crate::git::GitBrowseEntry;
use crate::gitutils::gitcommit::GitRef;
use crate::gitutils::gitdir::GitDir;
use crate::gitutils::gitfile::GitFile;
use std::path::{PathBuf, Path};
use std::str::FromStr;

pub struct GitRepo {
    git2 : Repository,
}

impl GitRepo {

    pub fn new(path : &str) -> Result<GitRepo, git2::Error> {
        Repository::open(path).map(|git2| GitRepo{git2})
    }

    pub fn get_tree<'a, 'b, 'c> (&'a self, commit : &'b GitRef, path : &'c Path) -> Result<GitDir<'a>, git2::Error> {
        let commit = match commit {
            GitRef::Commit(commit) => {
                // println!("commit : {}", commit.as_str());
                let oid = Oid::from_str(commit.as_str())?;
                // println!("{}", oid);
                self.git2.find_commit(oid)?
            }
            GitRef::Branch(branch) => {
                let branch = self.git2.find_branch(branch.as_str(), BranchType::Local)?;
                branch.get().peel_to_commit()?
            }
            GitRef::Head => {
                // println!("coucou1");
                let head = self.git2.head()?;
                // println!("coucou");
                head.peel_to_commit()?
            }
        };
        let root_tree = commit.tree()?;
        let tree = if path.as_os_str().len() == 0 {
            root_tree
        } else {
            let oid1 = root_tree.get_path(path)?.id();
            // println!("{}", oid1);
            self.git2.find_tree(oid1)?
        };
        let fullname = path.to_path_buf();
        Ok(GitDir { fullname, tree })
    }

    pub async fn browse<'a>(&'a self, dir: &GitDir<'a>) -> Result<Vec<GitBrowseEntry<'a>>, git2::Error> {
        let mut res = Vec::new();
        for entry in dir.tree.iter() {
            match entry.kind() {
                None => {Err(git2::Error::from_str("each tree entry must be well defined"))?;}
                Some(kind) => match kind {
                    ObjectType::Any => {Err(git2::Error::from_str("tree entry cannot be of kind Any"))?;}
                    ObjectType::Commit => {Err(git2::Error::from_str("tree entry cannot be of kind Commit"))?;}
                    ObjectType::Tree => {
                        let name = entry.name().ok_or(git2::Error::from_str("entry must have valid utf8 name"))?;
                        let mut fullname = dir.fullname.clone();
                        fullname.push(name);
                        let subtree = self.git2.find_tree(entry.id())?;
                        res.push(GitBrowseEntry::EGitDir(GitDir {fullname, tree : subtree}));
                    }
                    ObjectType::Blob => {
                        let name = entry.name().ok_or(git2::Error::from_str("entry must have valid utf8 name"))?;
                        let mut fullname = dir.fullname.clone();
                        fullname.push(name);
                        res.push(GitBrowseEntry::EGitFile(GitFile::new(fullname)));
                    }
                    ObjectType::Tag => {Err(git2::Error::from_str("tree entry cannot be of kind tag"))?;}
                }
            };
        }
        return Ok(res);
    }

    pub fn get_revisions_nbr(&self) -> u32 {
        let mut res = 0;
        let mut revwalk = self.git2.revwalk().expect("revisions walk is not defined");
        revwalk.push_head(); // todo : voir ce qu'il faut réellement pousser si on veux avoir TOUS les commits (genre par exemple si le HEAD n'est pas fils de tous les commits)
        for _ in revwalk{
            res = res + 1;
        }
        res
    }

    pub fn get_branches_nbr(&self) -> u32 {
        let mut res = 0;
        for _ in self.git2.branches(Some(BranchType::Local)){
            res = res + 1;
        }
        res
    }

    pub fn get_tags_nbr(&self) -> u32 {
        let mut res = 0;
        self.git2.tag_foreach(|_,_| {res = res + 1;true});
        res
    }

}
